import os, shutil, threading, subprocess
import dna.utils as utils
from dna.socat import SocatHelper


class DNA:
    """This is the main DNA class, which does all the heavy lifting

    :param service_name: the name of this DNA instance (up to you)
    :type service_name: str
    :param default: the default domain this DNA instance runs on (used by `nginx`)
    :type default: str
    :param cb_args: additional arguments to be used whenever `certbot` is called
    :type cb_args: list[str]

    When a DNA instance is initialized, the following occurs:

    * :meth:`~dna.DNA.configure` is called with the `service_name` parameter
    * an :class:`~dna.utils.Nginx` instance is created with the `default` parameter
    * a :class:`~dna.utils.Docker` instance is created
    * a :class:`~dna.utils.Certbot` instance is created with the `cb_args`\
        parameter, and its installer is set to nginx
    * a :class:`~dna.SocatHelper` instance is created with the `service_name`\
        parameter, as well as the instance's Docker object and path generated by\
        :meth:`~dna.DNA.configure`
    * :meth:`~dna.DNA.propagate_services` is called
    * SocatHelper's :meth:`~dna.SocatHelper.bind_sockets` is called with the\
        instance's services list

    When initialized, calls :meth:`~dna.DNA.configure`.
    """

    def __init__(self, service_name, default=None, cb_args=[]):
        self.configure(service_name)

        self.nginx = utils.Nginx(default)
        self.docker = utils.Docker()
        self.certbot = utils.Certbot(cb_args + ["-i", "nginx"])
        self.socat = SocatHelper(service_name, self.docker, self.path)

        self.propagate_services()
        self.socat.bind_sockets(self.services)

    def _make_dir(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def configure(self, service_name):
        """Configures the DNA instance called `service_name` for use

        :param service_name: the name of the DNA instance
        :type service_name: str

        * Creates the `.dna` folder and relevant subfolders as needed
        * Modify nginx to include configs made under this DNA instance
        * Creates a :class:`~dna.utils.SQLite` database for this DNA instance
        """
        self.service_name = service_name
        self.path = os.getcwd() + "/.dna"
        self.socks = self.path + "/socks"
        self.confs = self.path + "/nginx"
        self.logs = self.path + "/logs"

        for path in [self.path, self.socks, self.confs, self.logs]:
            self._make_dir(path)
        with open(f"/etc/nginx/conf.d/{self.service_name}.conf", "w") as nconf:
            nconf.write(f"include {self.confs}/*.conf;")

        self.db = utils.SQLite(rel="/.dna/", name=service_name)

    def get_service_info(self, service):
        """Gets the requested service

        :param service: the name of the service to find
        :type service: str

        :return: a :class:`~dna.utils.Service` object\
            representing the requested service
        """
        return self.db.get_service_by_name(service)

    def propagate_services(self):
        """Populates `self.services` with the services managed by this DNA\
            instance
        
        For the names of all the docker containers connected to this\
            service's socat bridge, calls :meth:`~dna.DNA.get_service_info`
        """
        dna = self.docker.get_network(self.socat.bridge, low_level=True)
        self.services = []
        for con in dna["Containers"]:
            if dna["Containers"][con]["Name"] == self.socat.container:
                continue
            service = self.get_service_info(dna["Containers"][con]["Name"])
            if not service:
                continue
            self.services.append(service)

    def do_docker_deploy(self, service, image, **options):
        """Deploys the image named `image` to a container named `service`

        :param service: the name of the service being launched
        :type service: str
        :param image: the name of the image holding the service
        :type image: str
        :param options: other options to pass to docker on deploy
        :type options: kwargs
        """
        print("Finding and killing container, if it exists...")
        self.docker.wipe_container(service)

        print("Starting container...")
        con = self.docker.run_image(
            image, service, detach=True, network=self.socat.bridge, **options
        )

        print("Pruning images...")
        self.docker.prune_images()

        print(f"Done! Successfully deployed {image} as {service}.")

    def do_nginx_deploy(self, service, domain):
        socket = f"{self.socks}/{service}.sock"
        with open(f"{self.confs}/{domain}.conf", "w") as out:
            out.write(
                self.nginx.gen_config_with_sock(
                    domain, socket, logs_pre=f"{self.logs}/{service}-"
                )
            )
        subprocess.call(["nginx", "-s", "reload"])

        cert = self.certbot.cert_else_false(domain)
        if cert:
            self.certbot.attach_cert(cert, domain)
        else:
            self.certbot.run_bot([domain])

    def do_db_deploy(self, service, image, port, domain):
        print("Doing database deploy...")
        if not self.db.get_service_by_name(service):
            self.db.create_service(service, image, port, [domain])
        print("Done!")

    def run_deploy(
        self, service, image, port, domain, threaded=False, **docker_options
    ):
        def threaded_deploy():
            self.do_docker_deploy(service, image, **docker_options)
            self.socat.bind(service, port)
            self.do_nginx_deploy(service, domain)

        if threaded:
            DockerDeploy(threaded_deploy).start()
        else:
            threaded_deploy()

        self.do_db_deploy(service, image, port, domain)
        self.propagate_services()

    def add_domain(self, service, domain):
        if self.db.add_domain_to_service(domain, service):
            self.do_nginx_deploy(service, domain)
        self.propagate_services()

    def pull_image(self, image, tag=None):
        self.docker.pull_image(image, tag)

    def build_image(self, **options):
        self.docker.build_image(**options)

    def delete_service(self, service):
        service = self.db.get_service_by_name(service)

        if not service:
            return

        for domain in service.domains:
            os.remove(f"{self.confs}/{domain.url}.conf")
        subprocess.call(["nginx", "-s", "reload"])

        self.socat.unbind(service.name, service.port)
        self.docker.wipe_container(service.name)

        self.db.delete_service(service)
        self.propagate_services()

    def docker_logs(self, service):
        return self.docker.service_logs(service)

    def nginx_logs(self, service, error=False):
        path = f'{self.logs}/{service}-{"error" if error else "access"}.log'
        with open(path) as f:
            return f.read()


class DockerDeploy(threading.Thread):
    def __init__(self, runnable):
        threading.Thread.__init__(self)
        self.runnable = runnable
        self.daemon = True

    def run(self):
        self.runnable()
