from sqlalchemy import Column, String, ForeignKey, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, relationship, backref

Base = declarative_base()


class Domain(Base):
    __tablename__ = "domain"
    url = Column(String, primary_key=True)
    service_name = Column(String, ForeignKey("service.name"))

    def __repr__(self):
        return f"Domain({self.url}" + (
            f", {self.service_name})" if self.service_name else ")"
        )


class Service(Base):
    __tablename__ = "service"
    name = Column(String, primary_key=True)
    image = Column(String)
    port = Column(String)
    domains = relationship("Domain", backref=backref("service"))

    def __repr__(self):
        res = f"Service({self.name}, {self.image}, {self.port}"
        for d in self.domains:
            res += f", {d.url}"
        return res + ")"


class SQLite:
    def __init__(self, rel="/", name="app"):
        engine = create_engine(f"sqlite://{rel}{name}.db?check_same_thread=False")
        Base.metadata.create_all(engine)
        self.s = Session(engine)

    def get_services(self):
        return self.s.query(Service).all()

    def get_service_by_name(self, name):
        return self.s.query(Service).filter(Service.name == name).one_or_none()

    def get_service_by_port(self, port):
        return self.s.query(Service).filter(Service.port == port).one_or_none()

    def get_service_by_domain(self, domain):
        if isinstance(domain, str):
            return (
                self.s.query(Service)
                .filter(any(d.url == domain for d in Service.domains))
                .one_or_none()
            )
        return (
            self.s.query(Service)
            .filte(any(d == domain for d in Service.domains))
            .one_or_none()
        )

    def add_service(self, service):
        self.s.add(service)
        self.s.commit()

    def create_service(self, name, image, port, domains=[]):
        s = Service(name=name, image=image, port=port)
        for d in domains:
            d = self.get_or_create_domain(d)
            if d.service_name:
                print(
                    f"Skipping {d.url} because it is already mapped to {d.service_name}."
                )
            else:
                s.domains.append(d)
        self.add_service(s)

    def add_domain_to_service(self, domain, service):
        if isinstance(service, str):
            service = self.get_service_by_name(service)
        if isinstance(domain, str):
            domain = self.get_or_create_domain(domain)

        if domain.service:
            return domain.service == service

        service.domains.append(domain)
        self.s.commit()

        return True

    def get_domains(self):
        return self.s.query(Domain).all()

    def get_domain_by_url(self, url):
        return self.s.query(Domain).filter(Domain.url == url).one_or_none()

    def get_domain_by_service(self, service):
        if isinstance(service, str):
            return (
                self.s.query(Domain)
                .filter(Domain.service_name == service)
                .one_or_none()
            )
        return self.s.query(Domain).filter(Domain.service == service).one_or_none()

    def get_or_create_domain(self, url):
        get = self.get_domain_by_url(url)
        if get:
            return get
        domain = Domain(url=url)
        self.add_domain(domain)
        return domain

    def add_domain(self, domain):
        self.s.add(domain)
        self.s.commit()

    def delete_service(self, service):
        if isinstance(service, str):
            service = self.get_service_by_name(service)
        for domain in service.domains:
            self.s.delete(domain)
        self.s.delete(service)
        self.s.commit()
